#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <semaphore.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>

#include <cam_common.h>
#include <cam_ipc.h>
#include <profile.h>

MODULE_DECLARE_MAGIC_STATIC;

#define STR_ERR_MAX_LEN 128
#define SET_NOERR(d) (d->errstr[0] = 0)
#define ERRSTR_UPDATE(desc) do { \
	snprintf(d->errstr, \
			STR_ERR_MAX_LEN, \
			"[%s] %s failed w/errno=%d (%s)", \
			__FUNCTION__, \
			desc, \
			errno, \
			strerror(errno)); \
} while (0)

typedef struct {
	sipc_mode_t mode;
	const char *addr;
	int connection_fd, accept_fd;
#ifdef CAMD_ON_DEMAND
	sem_t *sem;
#endif
	char errstr[STR_ERR_MAX_LEN];
	int bg_initctl_pid;
	const char *bg_initctl_desc;
	 MODULE_DECLARE_MAGIC_MEMBER;
} sipc_data_t;

sipc_data_t sipc_data_singleton;

size_t sipc_read(void *h, char *buf, size_t n)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	size_t pos = 0;
	ssize_t res;

	while (n > pos) {
		res = read(d->connection_fd, buf + pos, n - pos);
		switch ((int)res) {
		case -1:
			if (errno == EINTR || errno == EAGAIN)
				continue;
			return 0;
		case 0:
			errno = EPIPE;
			return pos;
		default:
			pos += (size_t) res;
		}
	}
	SET_NOERR(d);
	return pos;
}

size_t sipc_write(void *h, char *buf, size_t n)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	size_t pos = 0;
	ssize_t res;
	while (n > pos) {
		res = write(d->connection_fd, buf + pos, n - pos);
		switch ((int)res) {
		case -1:
			if (errno == EINTR || errno == EAGAIN)
				continue;
			return 0;
		case 0:
			errno = EPIPE;
			return pos;
		default:
			pos += (size_t) res;
		}
	}
	SET_NOERR(d);
	return pos;
}

static void sipc_set_abstract_ns(struct sockaddr_un *saun, const char *addr, int *len)
{
	memset(saun, 0, sizeof(struct sockaddr_un));
	saun->sun_family = AF_UNIX;

	/* leave one null byte at the beginning to indicate "abstract namespace" */
	strncpy(saun->sun_path + 1, addr, sizeof(saun->sun_path) - 1);

	/* add null terminator, in case addr is too long */
	saun->sun_path[sizeof(saun->sun_path) - 1] = 0;

	/* total length will be the total size of sockaddr_un - <unused addr bytes> */
	*len = sizeof(saun->sun_family) + 1 + strlen(addr);
}

static int sipc_bind(sipc_data_t *d)
{
	MODULE_VERIFY(d);
	int s, len, ret = -1;
	struct sockaddr_un saun;

	/*
	 * Get a socket to work with.  This socket will
	 * be in the UNIX domain, and will be a
	 * stream socket.
	 */
	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		ERRSTR_UPDATE("socket");
		goto sipc_bind_exit;
	}

	/*
	 * Create the address we will be binding to.
	 */
	sipc_set_abstract_ns(&saun, d->addr, &len);

	/*
	 * Try to bind the address to the socket. 	
	 * The third argument indicates the "length" of
	 * the structure, not just the length of the
	 * socket name.
	 */

	if (bind(s, (struct sockaddr *)&saun, len) < 0) {
		ERRSTR_UPDATE("bind");
		goto sipc_bind_exit;
	}

	/*
	 * Listen on the socket.
	 */
	if (listen(s, 5) < 0) {
		ERRSTR_UPDATE("listen");
		goto sipc_bind_exit;
	}
	d->accept_fd = s;
	SET_NOERR(d);
	ret = 0;
sipc_bind_exit:
	return ret;
}

int sipc_server_accept(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	socklen_t fromlen = sizeof(struct sockaddr_un);
	struct sockaddr_un fsaun;

	/*
	 * Accept connections.  When we accept one, connection_fd
	 * will be connected to the client.  fsaun will
	 * contain the address of the client.
	 */
	d->connection_fd = accept(d->accept_fd, (struct sockaddr *)&fsaun, &fromlen);

	if (d->connection_fd < 0) {
		ERRSTR_UPDATE("accept");
		goto sipc_server_accept_exit;
	}
	SET_NOERR(d);
sipc_server_accept_exit:
	return d->connection_fd != -1 ? 0 : -1;
}

#define DEFAULT_CONNECT_TIMEOUT_SEC 3

int sipc_client_connect_timeout(void *h, int timeout_sec)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	int s, len, ret = -1;
	struct sockaddr_un saun;

	/*
	 * Get a socket to work with.  This socket will
	 * be in the UNIX domain, and will be a
	 * stream socket.
	 */
	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		ERRSTR_UPDATE("socket");
		goto sipc_client_connect_exit;
	}

	/*
	 * Create the address we will be connecting to.
	 */
	sipc_set_abstract_ns(&saun, d->addr, &len);

#ifdef CAMD_ON_DEMAND
	/* attempt to wait for initctl and for the sem only if camd just started */
	if (d->bg_initctl_pid != -1) {
		int status;
		if (waitpid(d->bg_initctl_pid, &status, 0) != d->bg_initctl_pid) {
			ERRSTR_UPDATE("waitpid");
			goto sipc_client_connect_exit;
		}
		if (!WIFEXITED(status)) {
			char msg[STR_ERR_MAX_LEN];
			snprintf(msg, STR_ERR_MAX_LEN, "unable to initctl-%s camd",
					d->bg_initctl_desc);
			ERRSTR_UPDATE(msg);
			goto sipc_client_connect_exit;
		}
		d->bg_initctl_pid = -1;

		struct timespec abs_timeout;
		if (clock_gettime(CLOCK_REALTIME, &abs_timeout) == -1) {
			ERRSTR_UPDATE("clock_gettime");
			goto sipc_client_connect_exit;
		}
		if (timeout_sec == -1)
			timeout_sec = DEFAULT_CONNECT_TIMEOUT_SEC;
		abs_timeout.tv_sec += timeout_sec;

		if (sem_timedwait(d->sem, &abs_timeout)) {
			ERRSTR_UPDATE("sem_timedwait");
			goto sipc_client_connect_exit;
		}
	}
#endif		

	/*
	 * Try to connect to the address.  For this to
	 * succeed, the server must already have bound
	 * this address, and must have issued a listen()
	 * request.
	 *
	 * The third argument indicates the "length" of
	 * the structure, not just the length of the
	 * socket name.
	 */
	if (connect(s, (struct sockaddr *)&saun, len)) {
		ERRSTR_UPDATE("connect");
		goto sipc_client_connect_exit;
	}
	d->connection_fd = s;
	SET_NOERR(d);
	ret = 0;

sipc_client_connect_exit:
	if (ret) {
		if (d->connection_fd != -1 && close(d->connection_fd)) {
			/* nothing much to do, ret is already flagged */
		}

		d->connection_fd = -1;
		if (s != -1 && close(s)) {
			/* nothing much to do, ret is already flagged */
		}
	}
	return ret;
}

int sipc_disconnect(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = -1;

	if (d->connection_fd != -1) {
		if (close(d->connection_fd)) {
			ERRSTR_UPDATE("unable to close connection socket");
			goto sipc_disconnect_exit;
		}
		d->connection_fd = -1;
	}
	switch (d->mode) {
		case SIPC_MODE_SERVER:
			if (d->accept_fd != -1) {
				if (close(d->accept_fd)) {
					ERRSTR_UPDATE("unable to close accept socket");
					goto sipc_disconnect_exit;
				}
				d->accept_fd = -1;
			}
			break;
		case SIPC_MODE_CLIENT:
			break;
	}
	ret = 0;
sipc_disconnect_exit:
	return ret;
}

int sipc_get_connection_fd(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	return d->connection_fd;
}

void sipc_invalidate_connection_fd(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	d->connection_fd = -1;
}

int sipc_server_get_accept_fd(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	DEBUG_ASSERT(d->mode == SIPC_MODE_SERVER);
	return d->accept_fd;
}

void *sipc_create(void)
{
	sipc_data_t *d = &sipc_data_singleton;
	SET_NOERR(d);
	return d;
}

#ifdef CAMD_ON_DEMAND

#define INITCTL_BIN "/sbin/initctl"

static int initctl_camd(sipc_data_t *d, int enable, int wait_for_completion)
{
	int pid, ret = -1;

	if ((pid = fork()) == -1) {
		ERRSTR_UPDATE("fork");
		goto initctl_camd_exit;
	}

	if (pid == 0) {
		/* child */
		char **exec_argv = enable ?
			(char *[]){"start", "--no-wait", "camd", 0} :
			(char *[]){"stop", "camd", 0};
		if (execv(INITCTL_BIN, exec_argv) < 0)
			exit(1);
	}
	else {
		/* parent */
		const char *desc = enable ? "start" : "stop";
		if (wait_for_completion) {
			d->bg_initctl_pid = -1;
			d->bg_initctl_desc = NULL;
			int status;
			if (waitpid(pid, &status, 0) != pid) {
				ERRSTR_UPDATE("waitpid");
				goto initctl_camd_exit;
			}
			if (!WIFEXITED(status)) {
				char msg[STR_ERR_MAX_LEN];
				snprintf(msg, STR_ERR_MAX_LEN, 
						"unable to initctl-%s camd", desc);
				ERRSTR_UPDATE(msg);
				goto initctl_camd_exit;
			}
		}
		else {
			d->bg_initctl_pid = pid;
			d->bg_initctl_desc = desc;
		}
	}
	ret = 0;
initctl_camd_exit:
	return ret;
}
#endif

#define CAMD_SEM_NAME "camd"
#define CAMD_SEM_PATH "/dev/shm/sem." CAMD_SEM_NAME

int sipc_init(void *h, sipc_mode_t mode, const char *addr)
{
	sipc_data_t *d = h;
	MODULE_REGISTER(d, sipc_data_t);
	int ret = -1;

	d->addr = addr;
	d->mode = mode;

	switch (d->mode) {
		case SIPC_MODE_SERVER:

#ifdef CAMD_ON_DEMAND
			/* semaphore must exist */
			if ((d->sem = sem_open(CAMD_SEM_NAME, 0)) == SEM_FAILED) {
				ERRSTR_UPDATE("server sem_open");
				goto sipc_init_exit;
			}
#endif			
			if (sipc_bind(d))
				goto sipc_init_exit;
			/* client can connect now */
#ifdef CAMD_ON_DEMAND			
			if ((sem_post(d->sem))) {
				ERRSTR_UPDATE("server sem_post");
				goto sipc_init_exit;
			}
#endif			
			break;

		case SIPC_MODE_CLIENT:

#ifdef CAMD_ON_DEMAND
			PROF_SECT_DECL_BEGIN(stop_camd);

			/* stop camd */
			if (initctl_camd(d, 0, 1))
				goto sipc_init_exit;
			PROF_SECT_END(stop_camd);

			/* clean possible leftovers (crashes may leave the 
			 * semaphore up)
			 */
			struct stat stat_buf;
			if (!stat(CAMD_SEM_PATH, &stat_buf)) {
				snprintf(d->errstr, STR_ERR_MAX_LEN, 
						"%s already exists, client "
						"unclean exit detected",
						CAMD_SEM_PATH);
				if (sem_unlink(CAMD_SEM_NAME)) {
					ERRSTR_UPDATE("client sem_unlink");
					goto sipc_init_exit;
				}
			}
			if ((d->sem = sem_open(CAMD_SEM_NAME, O_CREAT | O_EXCL, 0x0666, 0)) 
					== SEM_FAILED) {
				ERRSTR_UPDATE("client sem_open");
				goto sipc_init_exit;
			}
			PROF_SECT_DECL_BEGIN(start_camd);

			/* start camd */
			if (initctl_camd(d, 1, 0))
				goto sipc_init_exit;
			PROF_SECT_END(start_camd);
#endif				
			break;
	}
	SET_NOERR(d);
	ret = 0;
sipc_init_exit:
	return ret;
}

int sipc_uninit(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);
	int ret = -1;

	/* disconnect, in case not explicitly called */
	if (sipc_disconnect(d))
		goto sipc_uninit_exit;

	switch (d->mode) {
		case SIPC_MODE_SERVER:
			break;
		case SIPC_MODE_CLIENT:
#ifdef CAMD_ON_DEMAND
			/* stop camd service
			 * not waiting for completion since start also stops 
			 * and waits 
			 */
			if (initctl_camd(d, 0, 1))
				goto sipc_uninit_exit;
#endif
			break;
	}

#ifdef CAMD_ON_DEMAND
	if (sem_close(d->sem)) {
		ERRSTR_UPDATE("sem_close");
		goto sipc_uninit_exit;
	}
	if (d->mode == SIPC_MODE_CLIENT) {
		if (sem_unlink(CAMD_SEM_NAME)) {
			ERRSTR_UPDATE("client sem_unlink");
			goto sipc_uninit_exit;
		}
	}
#endif
	d->addr = NULL;
	SET_NOERR(d);
	ret = 0;
sipc_uninit_exit:
	return ret;
}

const char *sipc_get_err_str(void *h)
{
	sipc_data_t *d = h;
	MODULE_VERIFY(d);

	return *(d->errstr) ? d->errstr : NULL;
}
