/* GStreamer
 *
 * Copyright (C) 2002 Ronald Bultje <rbultje@ronald.bitfreak.net>
 *               2006 Edgard Lima <edgard.lima@indt.org.br>
 *
 * v4l2src.c - system calls
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>
#include "v4l2src_calls.h"
#include <sys/time.h>
#include <unistd.h>
#include <poll.h>
#ifdef __sun
/* Needed on older Solaris Nevada builds (72 at least) */
#include <stropts.h>
#include <sys/ioccom.h>
#endif
/* for memalign */
#include <malloc.h>

/* On some systems MAP_FAILED seems to be missing */
#ifndef MAP_FAILED
#define MAP_FAILED ((caddr_t) -1)
#endif

/*** Still Capture Related ***/
#define DEFAULT_STILL_CAPTURE_FORMAT    V4L2_PIX_FMT_UYVY

#define GST_TYPE_V4L2_BUFFER (gst_v4l2_buffer_get_type())

static GstBufferClass *v4l2buffer_parent_class = NULL;

void gst_v4l2src_send_appmsg(GstV4l2Src * v4l2src, gint errCode)
{
	char *desc;

	switch (errCode) {
		case CAPTURE_START: 
			desc = "CaptureStart"; 
			break;
		case CAPTURE_SUCCESS: 
			desc = "CaptureComplete"; 
			break;
		default:
			g_assert(0);
	}

	GST_DEBUG("sending app msg %s", desc);
	GstMessage *msg = gst_message_new_application((GstObject *) v4l2src,
			gst_structure_new(desc, "Status", G_TYPE_INT, errCode, NULL));
	GstBus *bus = gst_element_get_bus((GstElement *) v4l2src);
	gst_bus_post(bus, msg);
	gst_object_unref(bus);
}

void still_capture_event_notify(GstV4l2Src * v4l2src)
{
	if (v4l2src->still_buffer_ready && 
		v4l2src->frame_create_mode == FCM_PREVIEW) {
		gst_v4l2src_send_appmsg(v4l2src, CAPTURE_SUCCESS);
		sem_post(&v4l2src->still_capture_sem);
	}
}

static void gst_v4l2_buffer_finalize(GstV4l2Buffer *buffer)
{
	gboolean resuscitated;

	GstV4l2Src *v4l2src = buffer->v4l2src;

	g_assert(GST_BUFFER_SIZE(buffer));
	GST_LOG("finalizing still cature buffer %p refount=%d", 
			buffer, GST_MINI_OBJECT_REFCOUNT_VALUE(buffer));

	g_mutex_lock(v4l2src->still_buffer_lock);

	if (v4l2src->relinquish_still_buffer) {
		GST_LOG("requeue not necessary");
		resuscitated = FALSE;
	} else {
		gst_buffer_ref(GST_BUFFER(buffer));
		v4l2src->still_capture_buffer = buffer;
		GST_DEBUG("still capture buffer refount=%d",
				GST_MINI_OBJECT_REFCOUNT_VALUE(
					v4l2src->still_capture_buffer));

		v4l2src->still_buffer_ready = TRUE;
		still_capture_event_notify(v4l2src);
		resuscitated = TRUE;
	}
	g_mutex_unlock(v4l2src->still_buffer_lock);
  
	if (!resuscitated) {
		/* unmap the v4l2 buffer at close */
		GST_LOG("still buffer %p not recovered, unmapping", buffer);
		ImageBuffer *ib = (ImageBuffer *)GST_BUFFER_DATA(buffer);
		gint ret = munmap(ib->buffer, buffer->vbuffer.length);
		g_assert(!ret);
		g_free(ib);
		GST_MINI_OBJECT_CLASS(v4l2buffer_parent_class)->finalize(
				GST_MINI_OBJECT(buffer));
	}
}

static void gst_v4l2_buffer_init(GstV4l2Buffer * xvimage, gpointer g_class)
{
	/* NOP */
}

static void gst_v4l2_buffer_class_init(gpointer g_class, gpointer class_data)
{
	GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS(g_class);

	v4l2buffer_parent_class = g_type_class_peek_parent(g_class);

	mini_object_class->finalize = (GstMiniObjectFinalizeFunction)
	    gst_v4l2_buffer_finalize;
}

static GType gst_v4l2_buffer_get_type(void)
{
	static GType _gst_v4l2_buffer_type;

	if (G_UNLIKELY(_gst_v4l2_buffer_type == 0)) {
		static const GTypeInfo v4l2_buffer_info = {
			sizeof(GstBufferClass),
			NULL,
			NULL,
			gst_v4l2_buffer_class_init,
			NULL,
			NULL,
			sizeof(GstV4l2Buffer),
			0,
			(GInstanceInitFunc) gst_v4l2_buffer_init,
			NULL
		};
		_gst_v4l2_buffer_type = g_type_register_static(GST_TYPE_BUFFER,
							       "GstV4l2Buffer",
							       &v4l2_buffer_info,
							       0);
	}
	return _gst_v4l2_buffer_type;
}

static GstV4l2Buffer *gst_v4l2_buffer_new(GstV4l2Src *v4l2src, GstCaps * caps)
{
	GstV4l2Buffer *ret;
	guint8 *data;
	int fd = v4l2src->v4l2object->still_fd;

	ret = (GstV4l2Buffer *) gst_mini_object_new(GST_TYPE_V4L2_BUFFER);

	ret->v4l2src = v4l2src;

	GST_LOG("creating buffer %p maped on fd=%d", ret, fd);

	memset(&ret->vbuffer, 0x00, sizeof(ret->vbuffer));
	ret->vbuffer.index = 0;
	ret->vbuffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	ret->vbuffer.memory = V4L2_MEMORY_MMAP;

	if (ioctl(fd, VIDIOC_QUERYBUF, &ret->vbuffer) < 0)
		goto querybuf_failed;

	GST_LOG("  index:     %u", ret->vbuffer.index);
	GST_LOG("  type:      %d", ret->vbuffer.type);
	GST_LOG("  bytesused: %u", ret->vbuffer.bytesused);
	GST_LOG("  flags:     %08x", ret->vbuffer.flags);
	GST_LOG("  field:     %d", ret->vbuffer.field);
	GST_LOG("  memory:    %d", ret->vbuffer.memory);
	GST_LOG("  MMAP offset:  %u", ret->vbuffer.m.offset);
	GST_LOG("  length:    %u", ret->vbuffer.length);
	GST_LOG("  input:     %u", ret->vbuffer.input);

	data = (guint8 *) mmap(0, ret->vbuffer.length,
			       PROT_READ | PROT_WRITE, MAP_SHARED,
			       fd, ret->vbuffer.m.offset);
	GST_LOG("mmaped still buffer (addr=%p)", data);

	if (data == MAP_FAILED)
		goto mmap_failed;

	ImageBuffer *ib;
	if (!(ib = (ImageBuffer *)g_malloc(sizeof(*ib))))
		goto no_mem;
	ib->buffer = data;
	GST_BUFFER_DATA(ret) = (void *)ib;
	GST_LOG("alloced container for still buffer (addr=%p)", ib);
	GST_BUFFER_SIZE(ret) = ret->vbuffer.length;
	GST_BUFFER_FLAG_SET(ret, GST_BUFFER_FLAG_READONLY);
	gst_buffer_set_caps(GST_BUFFER(ret), caps);

	return ret;

	/* ERRORS */
no_mem:
	{
		GST_WARNING("memory allocatiion failed");
		gst_buffer_unref(GST_BUFFER(ret));
		return NULL;
	}
querybuf_failed:
	{
		gint errnosave = errno;

		GST_WARNING("Failed QUERYBUF: %s", g_strerror(errnosave));
		gst_buffer_unref(GST_BUFFER(ret));
		errno = errnosave;
		return NULL;
	}
mmap_failed:
	{
		gint errnosave = errno;

		GST_WARNING("Failed to mmap: %s", g_strerror(errnosave));
		gst_buffer_unref(GST_BUFFER(ret));
		errno = errnosave;
		return NULL;
	}
}

gboolean queue_still_buffer(GstV4l2Src *v4l2src)
{
	gboolean ret = FALSE;

	struct v4l2_buffer *buf;

	buf = &v4l2src->still_capture_buffer->vbuffer;
	GST_LOG("queuing still buffer (addr,size)=(0x%p,%d)", buf, buf->length);
	PROF_SECT_DECL_BEGIN(still_qbuf);
	if (ioctl(v4l2src->v4l2object->still_fd, VIDIOC_QBUF, buf) < 0)
		goto queue_failed;
	PROF_SECT_END(still_qbuf);

#ifdef SHOW_FRAMES_QDQ
	GST_LOG("queued gst buffer %p", v4l2src->still_capture_buffer);
#endif
	ret = TRUE;
	goto done;

	/* ERRORS */
queue_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, FAILED,
				  (_("could not queue buffer ")),
				  ("failed: %s(errno=%d)", 
				   g_strerror(errno), errno));
		goto done;
	}
done:
	return ret;
}

/* complete made up ranking, the values themselves are meaningless */
#define YUV_BASE_RANK     1000
#define JPEG_BASE_RANK     500
#define DV_BASE_RANK       200
#define RGB_BASE_RANK      100
#define YUV_ODD_BASE_RANK   50
#define RGB_ODD_BASE_RANK   25
#define BAYER_BASE_RANK     15
#define GREY_BASE_RANK       5

static gint gst_v4l2src_format_get_rank(guint32 fourcc)
{
	switch (fourcc) {
	case V4L2_PIX_FMT_MJPEG:
		return JPEG_BASE_RANK;
	case V4L2_PIX_FMT_JPEG:
		return JPEG_BASE_RANK + 1;

	case V4L2_PIX_FMT_RGB332:
	case V4L2_PIX_FMT_RGB555:
	case V4L2_PIX_FMT_RGB555X:
	case V4L2_PIX_FMT_RGB565:
	case V4L2_PIX_FMT_RGB565X:
		return RGB_ODD_BASE_RANK;

	case V4L2_PIX_FMT_RGB24:
	case V4L2_PIX_FMT_BGR24:
		return RGB_BASE_RANK - 1;

	case V4L2_PIX_FMT_RGB32:
	case V4L2_PIX_FMT_BGR32:
		return RGB_BASE_RANK;

	case V4L2_PIX_FMT_GREY:	/* 8 Greyscale */
		return GREY_BASE_RANK;

	case V4L2_PIX_FMT_NV12:	/* 12 Y/CbCr 4:2:0 */
	case V4L2_PIX_FMT_NV21:	/* 12 Y/CrCb 4:2:0 */
	case V4L2_PIX_FMT_YYUV:	/* 16 YUV 4:2:2 */
	case V4L2_PIX_FMT_HI240:	/* 8 8-bit color */
		return YUV_ODD_BASE_RANK;

	case V4L2_PIX_FMT_YVU410:	/* YVU9, 9 bits per pixel */
		return YUV_BASE_RANK + 3;
	case V4L2_PIX_FMT_YUV410:	/* YUV9, 9 bits per pixel */
		return YUV_BASE_RANK + 2;
	case V4L2_PIX_FMT_YUV420:	/* I420, 12 bits per pixel */
		return YUV_BASE_RANK + 7;
	case V4L2_PIX_FMT_YUYV:	/* YUY2, 16 bits per pixel */
		return YUV_BASE_RANK + 10;
	case V4L2_PIX_FMT_YVU420:	/* YV12, 12 bits per pixel */
		return YUV_BASE_RANK + 6;
	case V4L2_PIX_FMT_UYVY:	/* UYVY, 16 bits per pixel */
		return YUV_BASE_RANK + 9;
	case V4L2_PIX_FMT_Y41P:	/* Y41P, 12 bits per pixel */
		return YUV_BASE_RANK + 5;
	case V4L2_PIX_FMT_YUV411P:	/* Y41B, 12 bits per pixel */
		return YUV_BASE_RANK + 4;
	case V4L2_PIX_FMT_YUV422P:	/* Y42B, 16 bits per pixel */
		return YUV_BASE_RANK + 8;

	case V4L2_PIX_FMT_DV:
		return DV_BASE_RANK;

	case V4L2_PIX_FMT_MPEG:	/* MPEG */
	case V4L2_PIX_FMT_WNVA:	/* Winnov hw compres */
		return 0;

	case V4L2_PIX_FMT_SBGGR8:
		return BAYER_BASE_RANK;

	default:
		break;
	}

	return 0;
}

static gint gst_v4l2src_format_cmp_func(gconstpointer a, gconstpointer b)
{
	guint32 pf1 = ((struct v4l2_fmtdesc *)a)->pixelformat;
	guint32 pf2 = ((struct v4l2_fmtdesc *)b)->pixelformat;

	if (pf1 == pf2)
		return 0;

	return gst_v4l2src_format_get_rank(pf2) -
	    gst_v4l2src_format_get_rank(pf1);
}

/******************************************************
 * gst_v4l2src_fill_format_list():
 *   create list of supported capture formats
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2src_fill_format_list(GstV4l2Src * v4l2src)
{
	gint n;
	struct v4l2_fmtdesc *format;

	GST_DEBUG("getting src format enumerations");

	/* format enumeration */
	for (n = 0;; n++) {
		format = g_new0(struct v4l2_fmtdesc, 1);

		format->index = n;
		format->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (ioctl
		    (v4l2src->v4l2object->video_fd, VIDIOC_ENUM_FMT,
		     format) < 0) {
			if (errno == EINVAL) {
				g_free(format);
				break;	/* end of enumeration */
			} else {
				goto failed;
			}
		}

		GST_LOG_OBJECT(v4l2src, "index:       %u", format->index);
		GST_LOG_OBJECT(v4l2src, "type:        %d", format->type);
		GST_LOG_OBJECT(v4l2src, "flags:       %08x", format->flags);
		GST_LOG_OBJECT(v4l2src, "description: '%s'", format->description);
		GST_LOG_OBJECT(v4l2src, "pixelformat: %" GST_FOURCC_FORMAT,
			       GST_FOURCC_ARGS(format->pixelformat));

		/* sort formats according to our preference; we do this,
		   because caps are probed in the order the formats are in the
		   list, and the order of formats in the final probed caps
		   matters for things like fixation */
		v4l2src->formats = g_slist_insert_sorted(v4l2src->formats, format,
				(GCompareFunc)gst_v4l2src_format_cmp_func);
	}

	GST_DEBUG("got %d format(s)", n);

	return TRUE;

	/* ERRORS */
      failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				  (_
				   ("Failed to enumerate possible video formats "
				    "device '%s' can work with"),
				   v4l2src->v4l2object->videodev),
				  ("Failed to get number %d in pixelformat "
				   "enumeration for %s. (%d - %s)",
				   n, v4l2src->v4l2object->videodev, errno,
				   g_strerror(errno)));
		g_free(format);
		return FALSE;
	}
}

/******************************************************
 * gst_v4l2src_clear_format_list():
 *   free list of supported capture formats
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2src_clear_format_list(GstV4l2Src * v4l2src)
{
	g_slist_foreach(v4l2src->formats, (GFunc) g_free, NULL);
	g_slist_free(v4l2src->formats);
	v4l2src->formats = NULL;

	return TRUE;
}

static void get_gst_relative_time(GstV4l2Src *v4l2src, GstClockTime *gst_timestamp_ref)
{
	/* must lock to get clock and base time */
	GST_OBJECT_LOCK(v4l2src);
	*gst_timestamp_ref = gst_clock_get_time(GST_ELEMENT_CLOCK(v4l2src)) - 
		GST_ELEMENT(v4l2src)->base_time;
	GST_OBJECT_UNLOCK(v4l2src);
}

GstFlowReturn gst_v4l2src_grab_still_frame(GstV4l2Src * v4l2src)
{
	struct v4l2_buffer buffer;

	while (1) {
		memset(&buffer, 0x00, sizeof(buffer));
		buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buffer.memory = V4L2_MEMORY_MMAP;

		if (!ioctl(v4l2src->v4l2object->still_fd, VIDIOC_DQBUF, &buffer)) {
			break;
		}

		if (errno == EINTR) {
			GST_DEBUG("interrupted, restarting syscall");
			continue;
		}

		goto grab_failed;
	}

	g_assert(buffer.index == 0);

	GST_LOG("DQ still buffer");

	GST_BUFFER_SIZE(v4l2src->still_capture_buffer) = buffer.bytesused;

	get_gst_relative_time(v4l2src, 
			&GST_BUFFER_TIMESTAMP(v4l2src->still_capture_buffer));

	GST_LOG("grabbed still frame %d flags %08x", buffer.sequence, buffer.flags);

	return GST_FLOW_OK;

	/* ERRORS */
grab_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, FAILED,
				(_("VIDIOC_DQBUF failed on device %s, errno=%d"), 
				 v4l2src->v4l2object->videodev, errno), 
				GST_ERROR_SYSTEM);
		return GST_FLOW_ERROR;
	}
}

GstFlowReturn gst_v4l2src_grab_video_frame(GstV4l2Src *v4l2src, GstBuffer **buf)
{
	struct v4l2_buffer buffer;

	while (1) {
		memset(&buffer, 0x00, sizeof(buffer));
		buffer.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buffer.memory = V4L2_MEMORY_MMAP;

		if (!ioctl(v4l2src->v4l2object->video_fd, VIDIOC_DQBUF, &buffer)) {
			break;
		}

		if (errno == EINTR) {
			GST_DEBUG("interrupted, restarting syscall");
			continue;
		}

		goto grab_failed;
	}

	v4l2src->num_queued_buffers--;
	v4l2src->dq_frame_count++;
	GstClockTime gst_ts_current_frame;

#define USE_KERNEL_TIMESTAMPS
#ifdef USE_KERNEL_TIMESTAMPS
	{
		/* use kernel timestamps of video buffers */
		struct timeval res;
		timersub(&buffer.timestamp, &v4l2src->tv_start, &res);
		gst_ts_current_frame = GST_TIMEVAL_TO_TIME(res);
	}
#ifdef SHOW_TIMESTAMPS_DELTA
	{
		GstClockTime gst_timestamp;

		get_gst_relative_time(v4l2src, &gst_timestamp);
		GstClockTime kernel_timestamp = gst_ts_current_frame;
		g_assert(gst_timestamp >= kernel_timestamp);
		GstClockTime diff = gst_timestamp - kernel_timestamp;
		int delta_ms = GST_TIME_AS_MSECONDS(diff);
		if (delta_ms > 1) {
			GST_DEBUG("frame timestamp is %lld gst-clock delta_ms=%d", 
					gst_ts_current_frame, delta_ms);
		}
	}
#endif
#else 
	/* timestamp frames using the gst pipeline main clock (less accurate) */
	get_gst_relative_time(v4l2src, &gst_ts_current_frame);
#endif
#ifdef WARN_FRAME_STARVATION
#define MAX_TS_GAP_MS 70
	{
		long long int ts_interval = gst_ts_current_frame - 
			v4l2src->last_video_buffer_ts;
		if (ts_interval > MAX_TS_GAP_MS * GST_MSECOND) {
			GST_WARNING("starvation detected between frames (%d,%d) "
					"gap=%lld ms", 
					v4l2src->dq_frame_count - 1, 
					v4l2src->dq_frame_count,
					GST_TIME_AS_MSECONDS(ts_interval));
		}
	}
#endif

#ifdef SHOW_FPS
#define FRAME_RATE_TRACK_INTERVAL_MS 2000
	GST_OBJECT_LOCK(v4l2src);
	GstClockTime now = gst_clock_get_time(GST_ELEMENT_CLOCK(v4l2src));
	GST_OBJECT_UNLOCK(v4l2src);

	if (v4l2src->frame_count_ts) {
		int interval_ms = GST_TIME_AS_MSECONDS(now - v4l2src->frame_count_ts);
		if (interval_ms >= FRAME_RATE_TRACK_INTERVAL_MS) {
			int frames_delta = v4l2src->dq_frame_count - 
				v4l2src->frame_count_last;
			// cheat by 1ms, makes FPS looks a bit better =)
			float fps = (frames_delta * 1000.0) / (interval_ms - 1);
			GST_INFO("streaming video @ %.2f fps", fps);
			v4l2src->frame_count_last = v4l2src->dq_frame_count;
			v4l2src->frame_count_ts = now;
		}
	}
	else
		v4l2src->frame_count_ts = now;
#endif
	/* that's the buffer we're retuning to the gst framework */
	*buf = v4l2src->video_buf[buffer.index].ptr;

	/* update timestamp */
	GST_BUFFER_TIMESTAMP(*buf) = gst_ts_current_frame;

	/* duration based on delta from previous frame, appears to be the most accurate */
	GST_BUFFER_DURATION(*buf) = gst_ts_current_frame - v4l2src->last_video_buffer_ts;
	v4l2src->last_video_buffer_ts = gst_ts_current_frame;

	/* TODO: why src file offests are necessary? */
	GST_BUFFER_OFFSET(*buf) = v4l2src->offset++;
	GST_BUFFER_OFFSET_END(*buf) = v4l2src->offset;
	v4l2src->video_buf[buffer.index].queued = FALSE;

#ifdef SHOW_FRAMES_QDQ
	GST_LOG("DQ %d index %d addr %p queued %d", 
			v4l2src->dq_frame_count, 
			buffer.index, 
			buffer.m.userptr, 
			v4l2src->num_queued_buffers);
#endif
	return GST_FLOW_OK;

	/* ERRORS */
grab_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, FAILED,
				(_("VIDIOC_DQBUF failed on device %s, errno=%d"), 
				 v4l2src->v4l2object->videodev, errno), 
				GST_ERROR_SYSTEM);
		return GST_FLOW_ERROR;
	}
}

static gboolean fractions_are_equal(gint num1, gint den1, gint num2, gint den2)
{
	GValue fraction1 = { 0, }, fraction2 = {
	0,};

	g_value_init(&fraction1, GST_TYPE_FRACTION);
	g_value_init(&fraction2, GST_TYPE_FRACTION);
	gst_value_set_fraction(&fraction1, num1, den1);
	gst_value_set_fraction(&fraction2, num2, den2);
	/* we know we don't have to unset the values in this case */
	return (gst_value_compare(&fraction1, &fraction2) == GST_VALUE_EQUAL);
}

gboolean gst_v4l2src_set_capture(GstV4l2Src * v4l2src, guint32 pixelformat,
			guint32 width, guint32 height, guint fps_n, guint fps_d)
{
	gint fd = v4l2src->v4l2object->video_fd;
	struct v4l2_format format;
	struct v4l2_streamparm stream;

	GST_DEBUG("setting video capture %dx%d, format %" GST_FOURCC_FORMAT, 
			width, height, GST_FOURCC_ARGS(pixelformat)); 
	GST_V4L2_CHECK_OPEN(v4l2src->v4l2object);
	GST_V4L2_CHECK_NOT_ACTIVE(v4l2src->v4l2object);

	/* note: order of VIDIOC_S_PARM and VIDIOC_S_FMT ioctls deliberately
	   changed, so that set frame rate will be called _after_ the next
	   resolution is set. */
	{
		memset(&format, 0x00, sizeof(struct v4l2_format));
		format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		if (ioctl(fd, VIDIOC_G_FMT, &format) < 0)
			goto get_fmt_failed;

		format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		format.fmt.pix.width = width;
		format.fmt.pix.height = height;
		format.fmt.pix.pixelformat = pixelformat;
		/* request whole frames; change when gstreamer supports
		   interlaced video */
		format.fmt.pix.field = V4L2_FIELD_INTERLACED;

		if (ioctl(fd, VIDIOC_S_FMT, &format) < 0)
			goto set_fmt_failed;
		if (format.fmt.pix.width != width
				|| format.fmt.pix.height != height)
			goto invalid_dimensions;

		if (format.fmt.pix.pixelformat != pixelformat)
			goto invalid_pixelformat;
	}
	{
		memset(&stream, 0x00, sizeof(struct v4l2_streamparm));
		stream.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		/* Note: V4L2 wants the frame interval, we have the frame rate */
		// Units of frame time is in 0.1usec
		stream.parm.capture.timeperframe.numerator = fps_d * 10000000;
		stream.parm.capture.timeperframe.denominator = fps_n;

		/* some cheap USB cam's won't accept any change */
		if (ioctl(fd, VIDIOC_S_PARM, &stream) < 0) {
			GST_ELEMENT_WARNING(v4l2src, RESOURCE, SETTINGS,
					(_("Video input device did not accept "
					   "new frame rate setting.")),
					GST_ERROR_SYSTEM);
			goto set_parm_failed;
		}

		v4l2src->fps_n = fps_n;
		v4l2src->fps_d = fps_d;
	}
	memset(&stream, 0x00, sizeof(struct v4l2_streamparm));
	stream.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(fd, VIDIOC_G_PARM, &stream) < 0) {
		GST_ELEMENT_WARNING(v4l2src, RESOURCE, SETTINGS,
				(_("Could not get parameters on device '%s'"),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		goto done;
	}

	/* Note: V4L2 provides the frame interval, we have the frame rate */
	if (fractions_are_equal(stream.parm.capture.timeperframe.numerator,
				stream.parm.capture.timeperframe.denominator,
				fps_d, fps_n)) {
		GST_LOG_OBJECT(v4l2src,
				"Desired framerate already set, nothing to do");
		goto done;
	}

done:

	return TRUE;

	/* ERRORS */
get_fmt_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_
				 ("Device '%s' does not support video capture"),
				 v4l2src->v4l2object->videodev),
				("Call to G_FMT failed: (%s)",
				 g_strerror(errno)));
		return FALSE;
	}
set_fmt_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_("Device '%s' cannot capture at %dx%d"),
				 v4l2src->v4l2object->videodev, width,
				 height),
				("Call to S_FMT failed for %"
				 GST_FOURCC_FORMAT " @ %dx%d: %s",
				 GST_FOURCC_ARGS(pixelformat), width, height,
				 g_strerror(errno)));
		return FALSE;
	}
invalid_dimensions:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_("Device '%s' cannot capture at %dx%d"),
				 v4l2src->v4l2object->videodev, width,
				 height),
				("Tried to capture at %dx%d, but device returned size %dx%d",
				 width, height, format.fmt.pix.width,
				 format.fmt.pix.height));
		return FALSE;
	}
invalid_pixelformat:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_
				 ("Device '%s' cannot capture in the specified format"),
				 v4l2src->v4l2object->videodev),
				("Tried to capture in %" GST_FOURCC_FORMAT
				 ", but device returned format" " %"
				 GST_FOURCC_FORMAT,
				 GST_FOURCC_ARGS(pixelformat),
				 GST_FOURCC_ARGS(format.fmt.pix.
					 pixelformat)));
		return FALSE;
	}
set_parm_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_("Could not set parameters on device '%s'"),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		return FALSE;
	}
}

gboolean still_capture_buff_init(GstV4l2Src * v4l2src, GstCaps *caps, gint fd)
{
	GST_DEBUG("initializing capture buffer");

	struct v4l2_requestbuffers breq;
	memset(&breq, 0, sizeof(struct v4l2_requestbuffers));

	breq.count = 1;
	breq.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	breq.memory = V4L2_MEMORY_MMAP;

	if (ioctl(fd, VIDIOC_REQBUFS, &breq) < 0)
		goto reqbufs_failed;

	GST_LOG_OBJECT(v4l2src, " count:  %u", breq.count);
	GST_LOG_OBJECT(v4l2src, " type:   %d", breq.type);
	GST_LOG_OBJECT(v4l2src, " memory: %d", breq.memory);

	g_assert(1 == breq.count);
	v4l2src->still_capture_buffer = gst_v4l2_buffer_new(v4l2src, caps);
	GST_DEBUG("still capture buffer refount: %d",
			GST_MINI_OBJECT_REFCOUNT_VALUE(v4l2src->still_capture_buffer));

	GST_INFO_OBJECT(v4l2src, "capturing buffers via mmap");
	return TRUE;

	/* ERRORS */
reqbufs_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, READ,
				(_("Could not get buffers from device '%s'."),
				 v4l2src->v4l2object->videodev),
				("error %d requesting buffer: %s",
				 errno, g_strerror(errno)));
		return FALSE;
	}
}

GstFlowReturn queue_available_preview_buffers(GstV4l2Src *v4l2src)
{
	GstFlowReturn ret = GST_FLOW_ERROR;
	GstPad	*srcpad = GST_BASE_SRC(v4l2src)->srcpad;
	g_assert(srcpad);
	GstBuffer *vs_buf = NULL;
	void *vs_buff_data;
	struct v4l2_buffer pvbuf = {}; /* preview v4l2 buffer */
	int i;

	pvbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	pvbuf.memory = V4L2_MEMORY_USERPTR;
	pvbuf.length = PREVIEW_FRAME_SIZE_BYTES;

	while (1) {
		GstFlowReturn gfr = gst_pad_alloc_buffer_and_set_caps(srcpad, 0,
						PREVIEW_FRAME_SIZE_BYTES,
						GST_PAD_CAPS(srcpad), &vs_buf);

#ifdef SHOW_FRAMES_QDQ
		GST_LOG("QDQ: gst_pad_alloc_buffer_and_set_caps ret=%d", gfr);
#endif
		if (gfr == GST_FLOW_RESEND) 
			break;
		if (gfr != GST_FLOW_OK) {
			GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
					(_("gst_pad_alloc_buffer_and_set_caps failed")),
					GST_ERROR_SYSTEM);
			goto queue_available_preview_buffers_exit;
		}
		g_assert(vs_buf);
		vs_buff_data = GST_BUFFER_DATA(vs_buf);

#ifdef SHOW_FRAMES_QDQ
		GST_LOG("QDQ: got gstbuffer addr %p data %p", vs_buf, vs_buff_data);
#endif
		/* find index of buffer */
		for (i = 0; i < NUM_OVERLAY_BUFFERS; i++) {
			if (v4l2src->video_buf[i].ptr == vs_buf) {
				pvbuf.index = i;
#define WARN_BUF_QUEUED_STR \
	"critical: pad_alloc returned buf %d while already queued in v4l2 driver"

				if (v4l2src->video_buf[i].queued) {
					GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
							(_(WARN_BUF_QUEUED_STR),
							 pvbuf.index), GST_ERROR_SYSTEM);
					goto queue_available_preview_buffers_exit;
				}
				break;
			}
		}

		if (i == NUM_OVERLAY_BUFFERS) {

			/* initial qing */
			for (i = 0; i < NUM_OVERLAY_BUFFERS; i++) {
				if (v4l2src->video_buf[i].ptr)
					continue;
				v4l2src->video_buf[i].ptr = vs_buf;
				pvbuf.index = i;
#ifdef SHOW_FRAMES_QDQ
				GST_LOG("Q: new buffer %p stored at index %d", vs_buf, i);
#endif
				break;
			}
			if (i == NUM_OVERLAY_BUFFERS) {
				GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
						(_("buffer not recognized and "
						"buffer list full")), GST_ERROR_SYSTEM);
				goto queue_available_preview_buffers_exit;
			}
		}

		/* data frame buffer */ 
		pvbuf.m.userptr = (unsigned long)vs_buff_data;

#ifdef DEBUG
		if (GST_BUFFER_SIZE(vs_buf) < PREVIEW_FRAME_SIZE_BYTES) {
			GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
					(_("gst_pad_alloc_buffer_and_set_caps returned"
					   "buffer size too small: %d"), 
					 GST_BUFFER_SIZE(vs_buf)),
					GST_ERROR_SYSTEM);
			goto queue_available_preview_buffers_exit;
		}
#endif
		v4l2src->video_buf[pvbuf.index].queued = TRUE;

		if (ioctl(v4l2src->v4l2object->video_fd, VIDIOC_QBUF, &pvbuf) < 0) {
			GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
					(_("could not queue buffer (data userptr=%p) "
					"@ index %d"), pvbuf.m.userptr, pvbuf.index),
					GST_ERROR_SYSTEM);
			goto queue_available_preview_buffers_exit;
		}

		v4l2src->num_queued_buffers++;
		if (!v4l2src->last_video_buffer_ts)
			get_gst_relative_time(v4l2src, &v4l2src->last_video_buffer_ts);

#ifdef SHOW_FRAMES_QDQ
		GST_DEBUG("Q: addr %p index %d queued %d", pvbuf.m.userptr, pvbuf.index, 
				v4l2src->num_queued_buffers);
#endif
	}
	ret = GST_FLOW_OK;
queue_available_preview_buffers_exit:
	return ret;
}

/******************************************************
 * gst_v4l2src_capture_start():
 *   start streaming capture
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2src_capture_start(GstV4l2Src *v4l2src)
{
	gint type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	gint fd = v4l2src->v4l2object->video_fd;
	gboolean ret = FALSE;

	GST_V4L2_CHECK_ACTIVE(v4l2src->v4l2object);

	GST_INFO_OBJECT(v4l2src, "about to start preview");

	PROF_SECT_DECL_BEGIN(wait_for_camd_to_cfg_preview);
	if (CAMD_NOTIFICATION_SEND(v4l2src, CAMD_MSG_ID_N_PREVIEW, CAMD_MSG_ARG_STATE_ON)) {
		goto streamon_failed;
	}
	PROF_SECT_END(wait_for_camd_to_cfg_preview);

	PROF_SECT_DECL_BEGIN(preview_streamon);
	if (ioctl(fd, VIDIOC_STREAMON, &type) < 0)
		goto streamon_failed;
	PROF_SECT_END(preview_streamon);

	GST_INFO_OBJECT(v4l2src, "preview started");
#ifdef SHOW_FPS
	v4l2src->frame_count_last = 0;
	v4l2src->frame_count_ts = 0;
	v4l2src->frame_count_last = v4l2src->dq_frame_count;
#endif
	v4l2src->last_video_buffer_ts = 0;
	v4l2src->num_queued_buffers = 0;
	v4l2src->dq_frame_count = 0;

	/* lock here is unnecessary (?) */
	GST_TIME_TO_TIMEVAL(GST_ELEMENT(v4l2src)->base_time, v4l2src->tv_start);

	ret = TRUE;
	goto gst_v4l2src_capture_start_exit;

streamon_failed:
	GST_ELEMENT_ERROR(v4l2src, RESOURCE, OPEN_READ,
			(_("Error starting streaming capture from device '%s'."),
			 v4l2src->v4l2object->videodev), GST_ERROR_SYSTEM);
	goto gst_v4l2src_capture_start_exit;

gst_v4l2src_capture_start_exit:
	return ret;
}

/******************************************************
 * gst_v4l2src_capture_stop():
 *   stop streaming capture
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2src_capture_stop(GstV4l2Src * v4l2src)
{
	gboolean ret = FALSE;

	if (!GST_V4L2_IS_OPEN(v4l2src->v4l2object) || 
			!GST_V4L2_IS_ACTIVE(v4l2src->v4l2object)) {
		goto done;
	}

	if (CAMD_NOTIFICATION_SEND(v4l2src,
				CAMD_MSG_ID_N_PREVIEW,
				CAMD_MSG_ARG_STATE_OFF)) {
		/* ignore this as an error object was already 
		 * generated and there isn't much we can do about it now
		 */
	}
	GST_LOG("issuing stream off");
	gint type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl (v4l2src->v4l2object->video_fd, VIDIOC_STREAMOFF, &type) < 0)
		goto streamoff_failed;

	/* unref the preview buffers, will ask for those on next preview */
	int i;
	for (i = 0; i < NUM_OVERLAY_BUFFERS; i++) {
		if (v4l2src->video_buf[i].queued) {
			gst_buffer_unref(v4l2src->video_buf[i].ptr);
			v4l2src->video_buf[i].ptr = NULL;
			v4l2src->video_buf[i].queued = FALSE;
		}
	}
	ret = TRUE;
	goto done;

	/* ERRORS */
streamoff_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				  (_("Error stopping streaming capture from device '%s'"),
				   v4l2src->v4l2object->videodev),
				  GST_ERROR_SYSTEM);
		goto done;
	}

done:
	return ret;
}

/******************************************************
 * gst_v4l2src_capture_deinit():
 *   deinitialize the capture system
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2src_capture_deinit(GstV4l2Src * v4l2src)
{
	GST_DEBUG("deinitting capture system");

	if (!GST_V4L2_IS_OPEN(v4l2src->v4l2object)) {
		return TRUE;
	}
	if (!GST_V4L2_IS_ACTIVE(v4l2src->v4l2object)) {
		return TRUE;
	}
	GST_V4L2_SET_INACTIVE(v4l2src->v4l2object);

	return TRUE;
}

/****** Still capture buffer *****/
/******************************************************
 * gst_v4l2src_get_still_capture():
 *   get capture parameters
 * return value: TRUE on success, FALSE on error
 ******************************************************/
gboolean gst_v4l2src_get_still_capture(GstV4l2Src * v4l2src)
{
	GST_V4L2_CHECK_OPEN(v4l2src->v4l2object);

	v4l2src->still_format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	if (ioctl (v4l2src->v4l2object->still_fd, VIDIOC_G_FMT,
	     &(v4l2src->still_format)) < 0)
		goto fmt_failed;

	GST_DEBUG("[Still Capture VIDIOC_G_FMT] width = %d height = %d type = %d",
	     v4l2src->still_format.fmt.pix.width,
	     v4l2src->still_format.fmt.pix.height, v4l2src->still_format.type);

	return TRUE;

	/* ERRORS */
      fmt_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				  (_
				   ("Failed querying in which video format "
				   "device '%s' is working with"),
				   v4l2src->v4l2object->videodev),
				  ("Failed VIDIOC_G_FMT for %s. (%d - %s)",
				   v4l2src->v4l2object->videodev, errno,
				   g_strerror(errno)));
		return FALSE;
	}
}

gboolean gst_v4l2src_set_still_capture(GstV4l2Src * v4l2src)
{
	gint width, height;

	if (!GST_V4L2_IS_OPEN(v4l2src->v4l2object))
		goto fd_not_open;

	/* the order of VIDIOC_S_PARM and VIDIOC_S_FMT ioctls 
	 * sets frame rate _after_ the next resolution is set
	 */
	{
		width = v4l2src->still_format.fmt.pix.width;
		height = v4l2src->still_format.fmt.pix.height;

		GST_DEBUG("setting still capture format to %dx%d",
			  v4l2src->still_format.fmt.pix.width,
			  v4l2src->still_format.fmt.pix.height);

		v4l2src->still_format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		v4l2src->still_format.fmt.pix.pixelformat = 
			DEFAULT_STILL_CAPTURE_FORMAT;

		if (ioctl(v4l2src->v4l2object->still_fd, VIDIOC_S_FMT, 
			&v4l2src->still_format) < 0)
			goto fmt_failed;
		GST_LOG("still capture format set");
	}
	{
		struct v4l2_streamparm parm;
		// Set high quality mode
		memset(&parm, 0, sizeof(parm));
		parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		parm.parm.capture.capturemode = V4L2_MODE_HIGHQUALITY;
		parm.parm.capture.timeperframe.denominator = 1;
		if (ioctl(v4l2src->v4l2object->still_fd, VIDIOC_S_PARM, &parm) < 0)
			goto setparam_failed;
	}

	if (width != v4l2src->still_format.fmt.pix.width ||
	    height != v4l2src->still_format.fmt.pix.height) {
		GST_WARNING("The closest size from %dx%d is %dx%d, "
				"for video format %s on device '%s'",
				width, height,
				v4l2src->still_format.fmt.pix.width,
				v4l2src->still_format.fmt.pix.height,
				"unknown", v4l2src->v4l2object->videodev);
	}

	/* update internal info, posted error */
	if (!gst_v4l2src_get_still_capture(v4l2src))
		goto fail;

	if (DEFAULT_STILL_CAPTURE_FORMAT != v4l2src->still_format.fmt.pix.pixelformat)
		goto pixfmt_failed;

	return TRUE;

	/* ERRORS */
fd_not_open:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, OPEN_READ,
				(_("device '%s' is not open"),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		return FALSE;
	}
setparam_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, CLOSE,
				(_
				 ("Error High Quality mode of the Device '%s'."),
				 v4l2src->v4l2object->videodev),
				GST_ERROR_SYSTEM);
		return FALSE;
	}
fmt_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_
				 ("Failed setting the video format for device '%s'"),
				 v4l2src->v4l2object->videodev),
				("Failed to set pixelformat to %s @ %dx%d "
				 "for device %s. (%d - %s)",
				 "V4L2_PIX_FMT_YUYV", width, height,
				 v4l2src->v4l2object->videodev, errno,
				 g_strerror(errno)));
		return FALSE;
	}
pixfmt_failed:
	{
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SETTINGS,
				(_("video pixelformat does not match, expected %d got %d"),
				 DEFAULT_STILL_CAPTURE_FORMAT, v4l2src->still_format.fmt.pix.pixelformat),
				GST_ERROR_SYSTEM);
		return FALSE;
	}
fail:
	{
		/* ERROR was posted */
		return FALSE;
	}
}

/* send notification to camd, return 0 on success */
int send_camd_notification(GstV4l2Src * v4l2src,
			   camd_msg_id_t id, camd_msg_arg_t a, 
			   void *ptr, unsigned int size,
			   const char *desc)
{
	int ret;

	g_mutex_lock(v4l2src->camd_comm_lock);
	g_assert((size && ptr) || (!size && !ptr));
	if (ptr) 
		ret = camd_msg_request_data(v4l2src->h_camd_msg, id, a, ptr, size);
	else {
		camd_msg_packet_t p;

		camd_msg_packet_init(&p);
		camd_msg_packet_refcount_inc(&p);
		p.ph.id = id;
		p.ph.arg = a;
		p.ph.ack = CAMD_MSG_ACK_REQUEST;
		ret = camd_msg_packet_write_wait_ack(v4l2src->h_camd_msg, &p);
		camd_msg_packet_refcount_dec(&p);
	}

	if (ret == -1) {
		GST_ELEMENT_ERROR(v4l2src, RESOURCE, SYNC,
				(_("Unable to send notification "
				   "'%s' to camd, reason: %s"),
				 desc, camd_msg_get_err_str(v4l2src->h_camd_msg)), 
				GST_ERROR_SYSTEM);
	}

	g_mutex_unlock(v4l2src->camd_comm_lock);
	return ret;
}
